@@ -103,6 +103,49 @@ vec_slice_fallback <- function(x, i) {
   vec_restore(out, x)
 }
 
+vec_slice_fallback_integer64 <- function(x, i) {
+  d <- vec_dim_n(x)
+
+  if (d == 2) {
+    out <- x[i, , drop = FALSE]
+  } else {
+    miss_args <- rep(list(missing_arg()), d - 1)
+    out <- eval_bare(expr(x[i, !!!miss_args, drop = FALSE]))
+  }
+
+  is_na <- is.na(i)
+
+  if (!any(is_na)) {
+    return(out)
+  }
+
+  if (d == 2) {
+    out[is_na,] <- bit64::NA_integer64_
+  } else {
+    eval_bare(expr(out[is_na, !!!miss_args] <- bit64::NA_integer64_))
+  }
+
+  out
+}
+
+# bit64::integer64() objects do not have support for `NA_integer_`
+# slicing. This manually replaces the garbage values that are created
+# any time a slice with `NA_integer_` is made.
+vec_slice_dispatch_integer64 <- function(x, i) {
+  out <- x[i]
+
+  is_na <- is.na(i)
+
+  if (!any(is_na)) {
+    return(out)
+  }
+
+  out[is_na] <- bit64::NA_integer64_
+
+  out
+}
+
+
 #' @rdname vec_slice
 #' @export
 `vec_slice<-` <- function(x, i, value) {
@@ -162,34 +205,6 @@ vec_init <- function(x, n = 1L) {
   .Call(vctrs_init, x, n)
 }
 
-#' Repeatedly slice a vector
-#'
-#' `vec_chop()` provides an efficient method to repeatedly slice a vector. It
-#' captures the pattern of `map(indices, vec_slice, x = x)`.
-#'
-#' @param x A vector
-#' @param indices A list of index values to slice `x` with, or `NULL`. Each
-#'   element of the list must be an integer, character or logical vector that
-#'   would be valid as an index in [vec_slice()]. If `NULL`, `x` is split into
-#'   its individual elements, equivalent to using an `indices` of
-#'   `as.list(vec_seq_along(x))`.
-#' @return A list of size `vec_size(indices)` or, if `indices == NULL`,
-#'   `vec_size(x)`.
-#' @export
-#' @examples
-#' vec_chop(1:5)
-#' vec_chop(1:5, list(1, 1:2))
-#' vec_chop(mtcars, list(1:3, 4:6))
-vec_chop <- function(x, indices = NULL) {
-  .Call(vctrs_chop, x, indices)
-}
-
-# Exposed for testing  (`starts` is 0-based)
-vec_chop_seq <- function(x, starts, sizes, increasings = TRUE) {
-  args <- vec_recycle_common(starts, sizes, increasings)
-  .Call(vctrs_chop_seq, x, args[[1]], args[[2]], args[[3]])
-}
-
 # Exposed for testing (`start` is 0-based)
 vec_slice_seq <- function(x, start, size, increasing = TRUE) {
   .Call(vctrs_slice_seq, x, start, size, increasing)
