@@ -11,29 +11,22 @@
 #' @param n Number of rows. If `NULL`, will be computed from the length of
 #'   the first element of `x`.
 #' @param ...,class Additional arguments for creating subclasses.
+#'   The `"names"` and `"row.names"` attributes override input in `x` and `n`,
+#'   respectively:
+#'
+#'   - `"names"` is used if provided, overriding existing names in `x`
+#'   - `"row.names"` is used if provided, if `n` is provided it must be
+#'     consistent.
+#'
 #' @export
 #' @keywords internal
 #' @examples
 #' new_data_frame(list(x = 1:10, y = 10:1))
-new_data_frame <- function(x = list(), n = NULL, ..., class = character()) {
-  stopifnot(is.list(x))
-  n <- n %||% df_size(x)
-  stopifnot(is.integer(n), length(n) == 1L)
-
-  # names() should always be a character vector, but we can't enforce that
-  # because as.data.frame() returns a data frame with NULL names to indicate
-  # that outer names should be used
-  if (length(x) == 0) {
-    names(x) <- character()
-  }
-
-  structure(
-    x,
-    ...,
-    class = c(class, "data.frame"),
-    row.names = .set_row_names(n)
-  )
+new_data_frame <- function(x = list(), n = NULL, ..., class = NULL) {
+  .External(vctrs_new_data_frame, x, n, class, ...)
 }
+new_data_frame <- fn_inline_formals(new_data_frame, "x")
+
 
 # Light weight constructor used for tests - avoids having to repeatedly do
 # stringsAsFactors = FALSE etc. Should not be used in internal code as is
@@ -70,14 +63,10 @@ vec_ptype_abbr.data.frame <- function(x, ...) {
   paste0("df", vec_ptype_shape(x))
 }
 
-#' @export
-vec_proxy.data.frame <- function(x, ...) {
-  x
-}
 #' @export
 vec_proxy_compare.data.frame <- function(x, ..., relax = FALSE) {
-  x[] <- lapply(x[], vec_proxy_compare_default, relax = TRUE)
-  x
+  out <- lapply(as.list(x), vec_proxy_compare, relax = TRUE)
+  new_data_frame(out, nrow(x))
 }
 
 
@@ -115,6 +104,11 @@ vec_cast.data.frame.data.frame <- function(x, to, ..., x_arg = "x", to_arg = "to
   .Call(vctrs_df_as_dataframe, x, to, x_arg, to_arg)
 }
 #' @export
+#' @method vec_cast.data.frame list
+vec_cast.data.frame.list <- function(x, to, ..., x_arg = "x", to_arg = "to") {
+  vec_list_cast(x, to, x_arg = x_arg, to_arg = to_arg)
+}
+#' @export
 #' @method vec_cast.data.frame default
 vec_cast.data.frame.default <- function(x, to, ..., x_arg = "x", to_arg = "to") {
   vec_default_cast(x, to, x_arg = x_arg, to_arg = to_arg)
@@ -122,7 +116,7 @@ vec_cast.data.frame.default <- function(x, to, ..., x_arg = "x", to_arg = "to")
 
 #' @export
 vec_restore.data.frame <- function(x, to, ..., n = NULL) {
-  .Call(vctrs_df_restore, x, to, n)
+  .Call(vctrs_bare_df_restore, x, to, n)
 }
 
 
@@ -156,6 +150,6 @@ df_lossy_cast <- function(out, x, to) {
     lossy = length(extra) > 0,
     locations = int(),
     details = inline_list("Dropped variables: ", extra, quote = "`"),
-    .subclass = "vctrs_error_cast_lossy_dropped"
+    class = "vctrs_error_cast_lossy_dropped"
   )
 }
