@@ -1,4 +1,5 @@
 #include "vctrs.h"
+#include "type-data-frame.h"
 #include "utils.h"
 
 // Initialised at load time
@@ -9,6 +10,7 @@ SEXP fns_vec_proxy_equal_dispatch = NULL;
 // Defined below
 SEXP vec_proxy_method(SEXP x);
 SEXP vec_proxy_invoke(SEXP x, SEXP method);
+static SEXP vec_proxy_unwrap(SEXP x);
 
 
 // [[ register(); include("vctrs.h") ]]
@@ -30,8 +32,30 @@ SEXP vec_proxy(SEXP x) {
 
 // [[ register(); include("vctrs.h") ]]
 SEXP vec_proxy_equal(SEXP x) {
-  return vec_proxy_recursive(x, vctrs_proxy_equal);
+  SEXP proxy = PROTECT(vec_proxy_recursive(x, vctrs_proxy_equal));
+
+  if (is_data_frame(proxy)) {
+    // Flatten df-cols so we don't have to recurse to work with data
+    // frame proxies
+    proxy = PROTECT(df_flatten(proxy));
+
+    // Unwrap data frames of size 1 since the data frame wrapper
+    // doesn't impact rowwise equality or identity
+    proxy = vec_proxy_unwrap(proxy);
+
+    UNPROTECT(1);
+  }
+
+  UNPROTECT(1);
+  return proxy;
 }
+static SEXP vec_proxy_unwrap(SEXP x) {
+  if (TYPEOF(x) == VECSXP && XLENGTH(x) == 1 && is_data_frame(x)) {
+    x = vec_proxy_unwrap(VECTOR_ELT(x, 0));
+  }
+  return x;
+}
+
 SEXP vec_proxy_equal_dispatch(SEXP x) {
   if (vec_typeof(x) == vctrs_type_s3) {
     return vctrs_dispatch1(syms_vec_proxy_equal_dispatch, fns_vec_proxy_equal_dispatch,
@@ -83,7 +107,7 @@ SEXP vctrs_proxy_recursive(SEXP x, SEXP kind_) {
 }
 
 SEXP vec_proxy_method(SEXP x) {
-  return s3_find_method("vec_proxy", x);
+  return s3_find_method("vec_proxy", x, vctrs_method_table);
 }
 
 // This should be faster than normal dispatch but also means that
