@@ -1,5 +1,6 @@
 #include "vctrs.h"
 #include "utils.h"
+#include "subscript.h"
 #include "subscript-loc.h"
 
 static SEXP int_invert_location(SEXP subscript, R_len_t n,
@@ -15,6 +16,8 @@ static void stop_subscript_oob_name(SEXP i, SEXP names,
                                     const struct vec_as_location_opts* opts);
 static void stop_location_negative(SEXP i,
                                    const struct vec_as_location_opts* opts);
+static void stop_location_zero(SEXP i,
+                               const struct vec_as_location_opts* opts);
 static void stop_indicator_size(SEXP i, SEXP n,
                                 const struct vec_as_location_opts* opts);
 static void stop_location_negative_missing(SEXP i,
@@ -54,7 +57,11 @@ static SEXP int_as_location(SEXP subscript, R_len_t n,
       }
 
       if (elt == 0) {
-        ++n_zero;
+        switch (opts->loc_zero) {
+        case LOC_ZERO_REMOVE: ++n_zero; break;
+        case LOC_ZERO_ERROR: stop_location_zero(subscript, opts);
+        case LOC_ZERO_IGNORE: break;
+        }
       } else if (abs(elt) > n) {
         if (opts->loc_oob == LOC_OOB_ERROR) {
           stop_subscript_oob_location(subscript, n, opts);
@@ -273,26 +280,50 @@ SEXP vec_as_location(SEXP subscript, R_len_t n, SEXP names) {
   return vec_as_location_opts(subscript,
                               n,
                               names,
-                              vec_as_location_default_opts);
+                              vec_as_location_default_opts,
+                              NULL);
 }
 
 SEXP vec_as_location_opts(SEXP subscript, R_len_t n, SEXP names,
-                          const struct vec_as_location_opts* opts) {
+                          const struct vec_as_location_opts* location_opts,
+                          const struct vec_as_subscript_opts* subscript_opts) {
 
   if (vec_dim_n(subscript) != 1) {
     Rf_errorcall(R_NilValue, "`i` must have one dimension, not %d.", vec_dim_n(subscript));
   }
 
-  switch (TYPEOF(subscript)) {
-  case NILSXP: return vctrs_shared_empty_int;
-  case INTSXP: return int_as_location(subscript, n, opts);
-  case REALSXP: return dbl_as_location(subscript, n, opts);
-  case LGLSXP: return lgl_as_location(subscript, n, opts);
-  case STRSXP: return chr_as_location(subscript, names, opts);
+  ERR err = NULL;
+  if (subscript_opts) {
+    subscript = vec_as_subscript_opts(subscript, subscript_opts, &err);
+  } else {
+    struct vec_as_subscript_opts default_subscript_opts = {
+      .logical = SUBSCRIPT_TYPE_ACTION_CAST,
+      .numeric = SUBSCRIPT_TYPE_ACTION_CAST,
+      .character = SUBSCRIPT_TYPE_ACTION_CAST,
+      .subscript_arg = location_opts->subscript_arg
+    };
+    subscript = vec_as_subscript_opts(subscript, &default_subscript_opts, &err);
+  }
+  PROTECT2(subscript, err);
 
+  if (err) {
+    r_cnd_signal(err);
+    never_reached("vec_as_location_opts");
+  }
+
+  SEXP out = R_NilValue;
+  switch (TYPEOF(subscript)) {
+  case NILSXP: out = vctrs_shared_empty_int; break;
+  case INTSXP: out = int_as_location(subscript, n, location_opts); break;
+  case REALSXP: out = dbl_as_location(subscript, n, location_opts); break;
+  case LGLSXP: out = lgl_as_location(subscript, n, location_opts); break;
+  case STRSXP: out = chr_as_location(subscript, names, location_opts); break;
   default: Rf_errorcall(R_NilValue, "`i` must be an integer, character, or logical vector, not a %s.",
                         Rf_type2char(TYPEOF(subscript)));
   }
+
+  UNPROTECT(2);
+  return out;
 }
 
 static void stop_subscript_arg_missing() {
@@ -304,6 +335,9 @@ static void stop_bad_negative() {
 static void stop_bad_oob() {
   Rf_errorcall(R_NilValue, "`oob` must be one of \"error\" or \"extend\".");
 }
+static void stop_bad_zero() {
+  Rf_errorcall(R_NilValue, "`zero` must be one of \"remove\", \"error\", or \"ignore\".");
+}
 
 static enum subscript_missing parse_subscript_arg_missing(SEXP x) {
   if (TYPEOF(x) != STRSXP || Rf_length(x) == 0) {
@@ -345,10 +379,25 @@ static enum num_as_location_loc_oob parse_loc_oob(SEXP x) {
 
   never_reached("stop_bad_oob");
 }
+static enum num_as_location_loc_zero parse_loc_zero(SEXP x) {
+  if (TYPEOF(x) != STRSXP || Rf_length(x) == 0) {
+    stop_bad_zero();
+  }
 
+  const char* str = CHAR(STRING_ELT(x, 0));
+
+  if (!strcmp(str, "remove")) return LOC_ZERO_REMOVE;
+  if (!strcmp(str, "error")) return LOC_ZERO_ERROR;
+  if (!strcmp(str, "ignore")) return LOC_ZERO_IGNORE;
+  stop_bad_zero();
+
+  never_reached("parse_loc_zero");
+}
+
+// [[ register() ]]
 SEXP vctrs_as_location(SEXP subscript, SEXP n_, SEXP names,
-                       SEXP loc_negative, SEXP loc_oob,
-                       SEXP missing, SEXP arg) {
+                       SEXP loc_negative, SEXP loc_oob, SEXP loc_zero,
+                       SEXP missing, SEXP arg_) {
   R_len_t n = 0;
 
   if (n_ == R_NilValue && TYPEOF(subscript) == STRSXP) {
@@ -367,15 +416,17 @@ SEXP vctrs_as_location(SEXP subscript, SEXP n_, SEXP names,
     UNPROTECT(1);
   }
 
+  struct vctrs_arg arg = vec_as_arg(arg_);
   struct vec_as_location_opts opts = {
     .action = SUBSCRIPT_ACTION_DEFAULT,
     .missing = parse_subscript_arg_missing(missing),
     .loc_negative = parse_loc_negative(loc_negative),
     .loc_oob = parse_loc_oob(loc_oob),
-    .subscript_arg = arg
+    .loc_zero = parse_loc_zero(loc_zero),
+    .subscript_arg = &arg
   };
 
-  return vec_as_location_opts(subscript, n, names, &opts);
+  return vec_as_location_opts(subscript, n, names, &opts, NULL);
 }
 
 static void stop_subscript_missing(SEXP i) {
@@ -387,18 +438,20 @@ static void stop_subscript_missing(SEXP i) {
 
 static void stop_location_negative_missing(SEXP i,
                                            const struct vec_as_location_opts* opts) {
+  SEXP arg = PROTECT(vctrs_arg(opts->subscript_arg));
   vctrs_eval_mask3(Rf_install("stop_location_negative_missing"),
                    syms_i, i,
-                   syms_subscript_arg, opts->subscript_arg,
+                   syms_subscript_arg, arg,
                    syms_subscript_action, get_opts_action(opts),
                    vctrs_ns_env);
   never_reached("stop_location_negative_missing");
 }
 static void stop_location_negative_positive(SEXP i,
                                             const struct vec_as_location_opts* opts) {
+  SEXP arg = PROTECT(vctrs_arg(opts->subscript_arg));
   vctrs_eval_mask3(Rf_install("stop_location_negative_positive"),
                    syms_i, i,
-                   syms_subscript_arg, opts->subscript_arg,
+                   syms_subscript_arg, arg,
                    syms_subscript_action, get_opts_action(opts),
                    vctrs_ns_env);
   never_reached("stop_location_negative_positive");
@@ -407,12 +460,13 @@ static void stop_location_negative_positive(SEXP i,
 static void stop_subscript_oob_location(SEXP i, R_len_t size,
                                         const struct vec_as_location_opts* opts) {
   SEXP size_obj = PROTECT(r_int(size));
+  SEXP arg = PROTECT(vctrs_arg(opts->subscript_arg));
   vctrs_eval_mask5(Rf_install("stop_subscript_oob"),
                    syms_i, i,
                    syms_subscript_type, chrs_numeric,
                    syms_size, size_obj,
                    syms_subscript_action, get_opts_action(opts),
-                   syms_subscript_arg, opts->subscript_arg,
+                   syms_subscript_arg, arg,
                    vctrs_ns_env);
 
   UNPROTECT(1);
@@ -420,33 +474,47 @@ static void stop_subscript_oob_location(SEXP i, R_len_t size,
 }
 static void stop_subscript_oob_name(SEXP i, SEXP names,
                                     const struct vec_as_location_opts* opts) {
+  SEXP arg = PROTECT(vctrs_arg(opts->subscript_arg));
   vctrs_eval_mask5(Rf_install("stop_subscript_oob"),
                    syms_i, i,
                    syms_subscript_type, chrs_character,
                    syms_names, names,
                    syms_subscript_action, get_opts_action(opts),
-                   syms_subscript_arg, opts->subscript_arg,
+                   syms_subscript_arg, arg,
                    vctrs_ns_env);
   never_reached("stop_subscript_oob_name");
 }
 
 static void stop_location_negative(SEXP i,
                                    const struct vec_as_location_opts* opts) {
+  SEXP arg = PROTECT(vctrs_arg(opts->subscript_arg));
   vctrs_eval_mask3(Rf_install("stop_location_negative"),
                    syms_i, i,
                    syms_subscript_action, get_opts_action(opts),
-                   syms_subscript_arg, opts->subscript_arg,
+                   syms_subscript_arg, arg,
                    vctrs_ns_env);
   never_reached("stop_location_negative");
 }
 
+static void stop_location_zero(SEXP i,
+                               const struct vec_as_location_opts* opts) {
+  SEXP arg = PROTECT(vctrs_arg(opts->subscript_arg));
+  vctrs_eval_mask3(Rf_install("stop_location_zero"),
+                   syms_i, i,
+                   syms_subscript_action, get_opts_action(opts),
+                   syms_subscript_arg, arg,
+                   vctrs_ns_env);
+  never_reached("stop_location_zero");
+}
+
 static void stop_indicator_size(SEXP i, SEXP n,
                                 const struct vec_as_location_opts* opts) {
+  SEXP arg = PROTECT(vctrs_arg(opts->subscript_arg));
   vctrs_eval_mask4(Rf_install("stop_indicator_size"),
                    syms_i, i,
                    syms_n, n,
                    syms_subscript_action, get_opts_action(opts),
-                   syms_subscript_arg, opts->subscript_arg,
+                   syms_subscript_arg, arg,
                    vctrs_ns_env);
   never_reached("stop_indicator_size");
 }
@@ -454,11 +522,12 @@ static void stop_indicator_size(SEXP i, SEXP n,
 static void stop_location_oob_non_consecutive(SEXP i, R_len_t size,
                                               const struct vec_as_location_opts* opts) {
   SEXP size_obj = PROTECT(r_int(size));
+  SEXP arg = PROTECT(vctrs_arg(opts->subscript_arg));
   vctrs_eval_mask4(Rf_install("stop_location_oob_non_consecutive"),
                    syms_i, i,
                    syms_size, size_obj,
                    syms_subscript_action, get_opts_action(opts),
-                   syms_subscript_arg, opts->subscript_arg,
+                   syms_subscript_arg, arg,
                    vctrs_ns_env);
 
   UNPROTECT(1);
@@ -473,12 +542,14 @@ void vctrs_init_subscript_loc(SEXP ns) {
   vec_as_location_default_opts_obj.action = SUBSCRIPT_ACTION_DEFAULT;
   vec_as_location_default_opts_obj.loc_negative = LOC_NEGATIVE_INVERT;
   vec_as_location_default_opts_obj.loc_oob = LOC_OOB_ERROR;
-  vec_as_location_default_opts_obj.subscript_arg = R_NilValue;
+  vec_as_location_default_opts_obj.loc_zero = LOC_ZERO_REMOVE;
+  vec_as_location_default_opts_obj.subscript_arg = NULL;
   vec_as_location_default_opts_obj.missing = SUBSCRIPT_MISSING_PROPAGATE;
 
   vec_as_location_default_assign_opts_obj.action = SUBSCRIPT_ACTION_ASSIGN;
   vec_as_location_default_assign_opts_obj.loc_negative = LOC_NEGATIVE_INVERT;
   vec_as_location_default_assign_opts_obj.loc_oob = LOC_OOB_ERROR;
-  vec_as_location_default_assign_opts_obj.subscript_arg = R_NilValue;
+  vec_as_location_default_assign_opts_obj.loc_zero = LOC_ZERO_REMOVE;
+  vec_as_location_default_assign_opts_obj.subscript_arg = NULL;
   vec_as_location_default_assign_opts_obj.missing = SUBSCRIPT_MISSING_PROPAGATE;
 }
