@@ -170,12 +170,14 @@ test_that("can repair names in `vec_rbind()` (#229)", {
 
   expect_named(vec_rbind(list(`_` = 1)), "_")
   expect_named(vec_rbind(list(`_` = 1), .name_repair = "universal"), c("._"))
+
+  expect_named(vec_rbind(list(a = 1, a = 2), .name_repair = ~ toupper(.)), c("A", "A"))
 })
 
 test_that("can construct an id column", {
   df <- data.frame(x = 1)
 
-  expect_named(vec_rbind(df, df, .names_to = "id"), c("x", "id"))
+  expect_named(vec_rbind(df, df, .names_to = "id"), c("id", "x"))
   expect_equal(vec_rbind(df, df, .names_to = "id")$id, c(1L, 2L))
 
   expect_equal(vec_rbind(a = df, b = df, .names_to = "id")$id, c("a", "b"))
@@ -187,6 +189,46 @@ test_that("vec_rbind() fails with arrays of dimensionality > 3", {
   expect_error(vec_rbind(array(NA, c(1, 1, 1))), "Can't bind arrays")
 })
 
+test_that("row names are preserved by vec_rbind()", {
+  df1 <- mtcars[1:3, ]
+  df2 <- mtcars[4:5, ]
+  expect_identical(vec_rbind(df1, df2), mtcars[1:5, ])
+
+  row.names(df2) <- NULL
+  out <- mtcars[1:5, ]
+  row.names(out) <- c(row.names(df1), "...4", "...5")
+  expect_identical(vec_rbind(df1, df2), out)
+})
+
+test_that("can assign row names in vec_rbind()", {
+  df1 <- mtcars[1:3, ]
+  df2 <- mtcars[4:5, ]
+
+  # Combination
+  out <- vec_rbind(foo = df1, df2)
+  exp <- mtcars[1:5, ]
+  row.names(exp) <- c(paste0("foo...", row.names(df1)), row.names(df2))
+  expect_identical(out, exp)
+
+  out <- vec_rbind(foo = df1, df2, .names_to = "id")
+  exp <- mtcars[1:5, ]
+  exp <- vec_cbind(id = c(rep("foo", 3), rep("", 2)), exp)
+  expect_identical(out, exp)
+
+  # Sequence
+  out <- vec_rbind(foo = unrownames(df1), df2, bar = unrownames(mtcars[6, ]))
+  exp <- mtcars[1:6, ]
+  row.names(exp) <- c(paste0("foo", 1:3), row.names(df2), "bar")
+  expect_identical(out, exp)
+
+  out <- vec_rbind(foo = unrownames(df1), df2, bar = unrownames(mtcars[6, ]), .names_to = "id")
+  exp <- mtcars[1:6, ]
+  exp <- vec_cbind(id = c(rep("foo", 3), rep("", 2), "bar"), exp)
+  row.names(exp) <- c(paste0("...", 1:3), row.names(df2), "...6")
+  expect_identical(out, exp)
+})
+
+
 # cols --------------------------------------------------------------------
 
 test_that("empty inputs give data frame", {
@@ -269,6 +311,7 @@ test_that("can repair names in `vec_cbind()` (#227)", {
   expect_named(vec_cbind(`_` = 1, .name_repair = "universal"), "._")
 
   expect_named(vec_cbind(a = 1, a = 2, .name_repair = "minimal"), c("a", "a"))
+  expect_named(vec_cbind(a = 1, a = 2, .name_repair = toupper), c("A", "A"))
 })
 
 test_that("can supply `.names_to` to `vec_rbind()` (#229)", {
@@ -280,7 +323,7 @@ test_that("can supply `.names_to` to `vec_rbind()` (#229)", {
 
   expect_identical(
     vec_rbind(a = x, b = y, .names_to = "quux"),
-    data_frame(foo = c(1L, 2L, 5L), bar = c(3L, 4L, 6L), quux = c("a", "a", "b"))
+    data_frame(quux = c("a", "a", "b"), foo = c(1L, 2L, 5L), bar = c(3L, 4L, 6L))
   )
   expect_identical(
     vec_rbind(a = x, b = y, .names_to = "foo"),
@@ -290,7 +333,7 @@ test_that("can supply `.names_to` to `vec_rbind()` (#229)", {
   # No names
   expect_identical(
     vec_rbind(x, y, .names_to = "quux"),
-    data_frame(foo = c(1L, 2L, 5L), bar = c(3L, 4L, 6L), quux = c(1L, 1L, 2L))
+    data_frame(quux = c(1L, 1L, 2L), foo = c(1L, 2L, 5L), bar = c(3L, 4L, 6L))
   )
   expect_identical(
     vec_rbind(x, y, .names_to = "foo"),
@@ -301,6 +344,20 @@ test_that("can supply `.names_to` to `vec_rbind()` (#229)", {
   expect_identical(vec_rbind(x, b = y, .names_to = "quux")$quux, c("", "", "b"))
 })
 
+test_that("can supply existing `.names_to`", {
+  x <- data.frame(a = 1, id = TRUE)
+  expect_identical(
+    vec_rbind(foo = x, bar = c(a = 2), .names_to = "id"),
+    data_frame(a = c(1, 2), id = c("foo", "bar"))
+  )
+
+  y <- data.frame(id = TRUE, a = 1)
+  expect_identical(
+    vec_rbind(foo = y, bar = c(a = 2), .names_to = "id"),
+    data_frame(id = c("foo", "bar"), a = c(1, 2))
+  )
+})
+
 test_that("vec_cbind() returns visibly (#452)", {
   # Shouldn't be needed once `check_unique` is implemented in C
   expect_visible(vctrs::vec_cbind(x = 1, .name_repair = "check_unique"))
@@ -356,3 +413,107 @@ test_that("vec_cbind() fails with arrays of dimensionality > 3", {
   expect_error(vec_cbind(a), "Can't bind arrays")
   expect_error(vec_cbind(x = a), "Can't bind arrays")
 })
+
+test_that("vec_rbind() consistently handles unnamed outputs", {
+  # Name repair of columns is a little weird but unclear we can do better
+  expect_identical(
+    vec_rbind(1, 2),
+    data.frame(...1 = c(1, 2))
+  )
+  expect_identical(
+    vec_rbind(1, 2, ...10 = 3),
+    data.frame(...1 = c(1, 2, 3), row.names = c("...1", "...2", "...3"))
+  )
+
+  expect_identical(
+    vec_rbind(a = 1, b = 2),
+    data.frame(...1 = c(1, 2), row.names = c("a", "b"))
+  )
+  expect_identical(
+    vec_rbind(c(a = 1), c(b = 2)),
+    data.frame(a = c(1, NA), b = c(NA, 2))
+  )
+})
+
+test_that("vec_cbind() consistently handles unnamed outputs", {
+  expect_identical(
+    vec_cbind(1, 2),
+    data.frame(...1 = 1, ...2 = 2)
+  )
+  expect_identical(
+    vec_cbind(1, 2, ...10 = 3),
+    data.frame(...1 = 1, ...2 = 2, ...3 = 3)
+  )
+  expect_identical(
+    vec_cbind(a = 1, b = 2),
+    data.frame(a = 1, b = 2)
+  )
+  expect_identical(
+    vec_cbind(c(a = 1), c(b = 2)),
+    new_data_frame(list(...1 = c(a = 1), ...2 = c(b = 2)))
+  )
+})
+
+test_that("rbind() and cbind() have informative outputs when repairing names", {
+  verify_output(test_path("output", "bind-name-repair.txt"), {
+    "# vec_rbind()"
+
+    "Suboptimal"
+    vec_rbind(1, 2)
+
+    "Suboptimal"
+    vec_rbind(1, 2, ...10 = 3)
+
+    vec_rbind(a = 1, b = 2)
+    vec_rbind(c(a = 1), c(b = 2))
+
+    "# vec_cbind()"
+    vec_cbind(1, 2)
+    vec_cbind(1, 2, ...10 = 3)
+    vec_cbind(a = 1, b = 2)
+    vec_cbind(c(a = 1), c(b = 2))
+  })
+})
+
+test_that("cbind() deals with row names", {
+  expect_identical(
+    vec_cbind(mtcars[1:3], foo = 1),
+    cbind(mtcars[1:3], foo = 1)
+  )
+  expect_identical(
+    vec_cbind(mtcars[1:3], mtcars[4]),
+    cbind(mtcars[1:3], mtcars[4])
+  )
+
+  out <- vec_cbind(
+    mtcars[1, 1, drop = FALSE],
+    unrownames(mtcars[1:3, 2, drop = FALSE])
+  )
+  exp <- mtcars[1:3, c(1, 2)]
+  exp[[1]] <- exp[[1, 1]]
+  row.names(exp) <- paste0(c("Mazda RX4..."), 1:3)
+  expect_identical(out, exp)
+
+  # Should work once we have frame prototyping
+  expect_error(
+    vec_cbind(mtcars[1:3], vec_slice(mtcars[4], nrow(mtcars):1)),
+    "different row names"
+  )
+})
+
+test_that("can rbind data frames with matrix columns (#625)", {
+  df <- tibble(x = 1:2, y = matrix(1:4, nrow = 2))
+  expect_identical(vec_rbind(df, df), vec_slice(df, c(1, 2, 1, 2)))
+})
+
+test_that("rbind repairs names of data frames (#704)", {
+  df <- data_frame(x = 1, x = 2)
+  df_repaired <- data_frame(x...1 = 1, x...2 = 2)
+  expect_identical(vec_rbind(df), df_repaired)
+  expect_identical(vec_rbind(df, df), vec_rbind(df_repaired, df_repaired))
+
+  expect_error(
+    vec_rbind(df, df, .name_repair = "check_unique"),
+    class = "vctrs_error_names_must_be_unique"
+  )
+})
