@@ -83,13 +83,25 @@ test_that("can assign lists", {
 
 test_that("atomics can't be assigned in lists", {
   x <- list(NULL)
-  expect_error(vec_slice(x, 1) <- NA, class = "vctrs_error_incompatible_type")
-  expect_error(vec_assign(x, 1, NA), class = "vctrs_error_incompatible_type")
+  expect_error(vec_slice(x, 1) <- 1, class = "vctrs_error_incompatible_type")
+  expect_error(vec_assign(x, 1, 2), class = "vctrs_error_incompatible_type")
 
   expect_error(vec_slice(x, 1) <- "foo", class = "vctrs_error_incompatible_type")
   expect_error(vec_assign(x, 1, "foo"), class = "vctrs_error_incompatible_type")
 })
 
+test_that("Unspecified `NA` vector can be assigned into lists", {
+  x <- list(1, 2)
+  vec_slice(x, 1) <- NA
+  expect_identical(x, list(NULL, 2))
+})
+
+test_that("monitoring test - unspecified() can be assigned in lists", {
+  x <- list(1, 2)
+  expect_error(vec_slice(x, 1) <- unspecified(1), NA)
+  expect_equal(x, list(NULL, 2))
+})
+
 test_that("can assign and slice-assign data frames", {
   df <- data.frame(x = 1:2)
   df$y <- data.frame(a = 2:1)
@@ -260,7 +272,7 @@ test_that("can use names to vec_slice<-() a named object", {
 
 test_that("slice-assign falls back to `[<-` when proxy is not implemented", {
   obj <- foobar(c("foo", "bar", "baz"))
-  expect_error(vec_slice(obj, 1:2) <- NA, class = "vctrs_error_incompatible_cast")
+  expect_error(vec_slice(obj, 1:2) <- TRUE, class = "vctrs_error_incompatible_type")
 
   vec_slice(obj, 1:2) <- foobar("quux")
 
@@ -276,25 +288,32 @@ test_that("slice-assign falls back to `[<-` when proxy is not implemented", {
     },
     vec_ptype2.logical.vctrs_foobar = function(...) foobar(""),
     vec_ptype2.vctrs_foobar = function(...) foobar(""),
-    vec_cast.vctrs_foobar = function(x, to, ...) x
+    vec_cast.vctrs_foobar = function(x, to, ...) foobar(rep("", length(x)))
   )
 
   obj <- foobar(c("foo", "bar", "baz"))
-  vec_slice(obj, 1:2) <- NA
-  expect_identical(obj, c("dispatched", "dispatched", "baz"))
+  vec_slice(obj, 1:2) <- TRUE
+  expect_identical(obj, foobar(c("dispatched", "dispatched", "baz")))
 })
 
-test_that("slice-assign restores value before falling back to `[<-` (#443)", {
+test_that("vec_assign() can always assign unspecified values into foreign vector types", {
+  obj <- foobar(c("foo", "bar", "baz"))
+  expect <- foobar(c(NA, "bar", "baz"))
+
+  expect_identical(vec_assign(obj, 1, NA), expect)
+  expect_identical(vec_assign(obj, 1, unspecified(1)), expect)
+})
+
+test_that("slice-assign casts to `to` before falling back to `[<-` (#443)", {
   called <- FALSE
 
   local_methods(
     vec_proxy.vctrs_proxy = proxy_deref,
-    vec_restore.vctrs_proxy = function(x, to, ...) new_proxy(x),
     vec_ptype2.vctrs_proxy = function(...) new_proxy(NA),
-    vec_cast.vctrs_foobar = function(x, ...) proxy_deref(x),
+    vec_cast.vctrs_foobar = function(x, ...) foobar(proxy_deref(x)),
     `[<-.vctrs_foobar` = function(x, i, value) {
       called <<- TRUE
-      expect_is(value, "vctrs_proxy")
+      expect_identical(value, foobar(10))
     }
   )
 
@@ -318,6 +337,22 @@ test_that("can assign to data frame", {
   expect_identical(vec_assign(x, 2, y), data_frame(x = int(1, 20, 3)))
 })
 
+test_that("can assign to a data frame with matrix columns (#625)", {
+  df <- tibble(x = 1:2, y = matrix(1:4, nrow = 2))
+  expect_identical(vec_assign(df, 2L, df[1,]), vec_slice(df, c(1, 1)))
+})
+
+test_that("assigning to a factor doesn't produce corrupt levels (#853)", {
+  x <- factor(c("a", NA), levels = c("a", "b"))
+  value <- factor("b", levels = "b")
+
+  res <- vec_assign(x, 2, value)
+  expect_identical(res, factor(c("a", "b")))
+
+  res <- vec_assign(x, 1:2, value)
+  expect_identical(res, factor(c("b", "b"), levels = c("a", "b")))
+})
+
 test_that("can slice-assign unspecified vectors with default type2 method", {
   local_rational_class()
   x <- rational(1:2, 2:3)
