@@ -4,7 +4,7 @@ test_that("vec_slice throws error with non-vector inputs", {
 })
 
 test_that("vec_slice throws error with non-vector subscripts", {
-  expect_error(vec_slice(1:3, Sys.Date()), class = "vctrs_error_incompatible_cast")
+  expect_error(vec_slice(1:3, Sys.Date()), class = "vctrs_error_subscript_type")
   expect_error(vec_slice(1:3, matrix(TRUE, nrow = 1)), "must have one dimension")
 })
 
@@ -391,6 +391,48 @@ test_that("vec_slice() asserts vectorness (#301)", {
   expect_error(vec_slice(NULL, 1), class = "vctrs_error_scalar_type")
 })
 
+test_that("slicing an unspecified logical vector returns a logical vector", {
+  expect_identical(vec_slice(NA, integer()), logical())
+  expect_identical(vec_slice(NA, c(1, 1)), c(NA, NA))
+})
+
+test_that("slicing an unspecified() object returns an unspecified()", {
+  expect_identical(vec_slice(unspecified(1), integer()), unspecified())
+  expect_identical(vec_slice(unspecified(1), c(1, 1)), unspecified(2))
+})
+
+
+test_that("vec_slice() works with Altrep classes with custom extract methods", {
+  skip_if(getRversion() < "3.5")
+
+  x <- .Call(vctrs_rle, c(foo = 10L, bar = 5L))
+
+  idx <- c(9, 10, 11)
+  expect_equal(vec_slice(x, idx), c("foo", "foo", "bar"))
+})
+
+test_that("slice has informative error messages", {
+  verify_output(test_path("error", "test-slice.txt"), {
+    "# Unnamed vector with character subscript"
+    vec_slice(1:3, letters[1])
+
+    "# Negative subscripts are checked"
+    vec_slice(1:3, -c(1L, NA))
+    vec_slice(1:3, c(-1L, 1L))
+
+    "# oob error messages are properly constructed"
+    vec_slice(c(bar = 1), "foo")
+    "Multiple OOB indices"
+    vec_slice(letters, c(100, 1000))
+    vec_slice(letters, c(1, 100:103, 2, 104:110))
+    vec_slice(set_names(letters), c("foo", "bar"))
+    vec_slice(set_names(letters), toupper(letters))
+
+    "# Can't index beyond the end of a vector"
+    vec_slice(1:2, 3L)
+    vec_slice(1:2, -3L)
+  })
+})
 
 # vec_init ----------------------------------------------------------------
 
@@ -445,189 +487,6 @@ test_that("vec_init() works with Altrep classes", {
   expect_equal(vec_init(x, 2), rep(NA_character_, 2))
 })
 
-# vec_chop ----------------------------------------------------------------
-
-test_that("vec_chop() throws error with non-vector inputs", {
-  expect_error(vec_chop(NULL), class = "vctrs_error_scalar_type")
-  expect_error(vec_chop(environment()), class = "vctrs_error_scalar_type")
-})
-
-test_that("atomics are split into a list", {
-  x <- 1:5
-  expect_equal(vec_chop(x), as.list(x))
-
-  x <- letters[1:5]
-  expect_equal(vec_chop(x), as.list(x))
-})
-
-test_that("atomic names are kept", {
-  x <- set_names(1:5)
-  result <- lapply(vec_chop(x), names)
-  expect_equal(result, as.list(names(x)))
-})
-
-test_that("base R classed objects are split into a list", {
-  fctr <- factor(c("a", "b"))
-  expect <- lapply(vec_seq_along(fctr), vec_slice, x = fctr)
-  expect_equal(vec_chop(fctr), expect)
-
-  date <- new_date(c(0, 1))
-  expect <- lapply(vec_seq_along(date), vec_slice, x = date)
-  expect_equal(vec_chop(date), expect)
-})
-
-test_that("base R classed object names are kept", {
-  fctr <- set_names(factor(c("a", "b")))
-  result <- lapply(vec_chop(fctr), names)
-  expect_equal(result, as.list(names(fctr)))
-})
-
-test_that("list elements are split", {
-  x <- list(1, 2)
-  result <- list(vec_slice(x, 1), vec_slice(x, 2))
-  expect_equal(vec_chop(x), result)
-})
-
-test_that("data frames are split rowwise", {
-  x <- data_frame(x = 1:2, y = c("a", "b"))
-  result <- list(vec_slice(x, 1), vec_slice(x, 2))
-  expect_equal(vec_chop(x), result)
-})
-
-test_that("data frame row names are kept", {
-  x <- data_frame(x = 1:2, y = c("a", "b"))
-  rownames(x) <- c("r1", "r2")
-  result <- lapply(vec_chop(x), rownames)
-  expect_equal(result, list("r1", "r2"))
-})
-
-test_that("matrices / arrays are split rowwise", {
-  x <- array(1:12, c(2, 2, 2))
-  result <- list(vec_slice(x, 1), vec_slice(x, 2))
-  expect_equal(vec_chop(x), result)
-})
-
-test_that("matrix / array row names are kept", {
-  x <- array(1:12, c(2, 2, 2), dimnames = list(c("r1", "r2"), c("c1", "c2")))
-  result <- lapply(vec_chop(x), rownames)
-  expect_equal(result, list("r1", "r2"))
-})
-
-test_that("matrices / arrays without row names have other dimension names kept", {
-  x <- array(1:12, c(2, 2, 2), dimnames = list(NULL, c("c1", "c2")))
-  result <- lapply(vec_chop(x), colnames)
-  expect_equal(result, list(c("c1", "c2"), c("c1", "c2")))
-})
-
-test_that("vec_chop() doesn't restore when attributes have already been restored", {
-  local_methods(
-    `[.vctrs_foobar` = function(x, i, ...) structure("dispatched", foo = "bar"),
-    vec_restore.vctrs_foobar = function(...) structure("dispatched-and-restored", foo = "bar")
-  )
-
-  result <- vec_chop(foobar(NA))[[1]]
-  expect_equal(result, structure("dispatched", foo = "bar"))
-})
-
-test_that("vec_chop() restores when attributes have not been restored by `[`", {
-  local_methods(
-    `[.vctrs_foobar` = function(x, i, ...) "dispatched",
-    vec_restore.vctrs_foobar = function(...) "dispatched-and-restored"
-  )
-
-  result <- vec_chop(foobar(NA))[[1]]
-  expect_equal(result, "dispatched-and-restored")
-})
-
-test_that("vec_chop() falls back to `[` for shaped objects with no proxy", {
-  x <- foobar(1)
-  dim(x) <- c(1, 1)
-  result <- vec_chop(x)[[1]]
-  expect_equal(result, x)
-})
-
-test_that("`indices` are validated", {
-  expect_error(vec_chop(1, 1), "`indices` must be a list of index values, or `NULL`")
-  expect_error(vec_chop(1, list(1.5)), class = "vctrs_error_cast_lossy")
-  expect_error(vec_chop(1, list(2)), class = "vctrs_error_subscript_oob")
-})
-
-test_that("size 0 `indices` list is allowed", {
-  expect_equal(vec_chop(1, list()), list())
-})
-
-test_that("individual index values of size 0 are allowed", {
-  expect_equal(vec_chop(1, list(integer())), list(numeric()))
-
-  df <- data.frame(a = 1, b = "1")
-  expect_equal(vec_chop(df, list(integer())), list(vec_ptype(df)))
-})
-
-test_that("data frame row names are kept when `indices` are used", {
-  x <- data_frame(x = 1:2, y = c("a", "b"))
-  rownames(x) <- c("r1", "r2")
-  result <- lapply(vec_chop(x, list(1, 1:2)), rownames)
-  expect_equal(result, list("r1", c("r1", "r2")))
-})
-
-test_that("vec_chop(<atomic>, indices =) can be equivalent to the default", {
-  x <- 1:5
-  indices <- as.list(vec_seq_along(x))
-  expect_equal(vec_chop(x, indices), vec_chop(x))
-})
-
-test_that("vec_chop(<data.frame>, indices =) can be equivalent to the default", {
-  x <- data.frame(x = 1:5)
-  indices <- as.list(vec_seq_along(x))
-  expect_equal(vec_chop(x, indices), vec_chop(x))
-})
-
-test_that("vec_chop(<array>, indices =) can be equivalent to the default", {
-  x <- array(1:8, c(2, 2, 2))
-  indices <- as.list(vec_seq_along(x))
-  expect_equal(vec_chop(x, indices), vec_chop(x))
-})
-
-test_that("`indices` can use names", {
-  x <- set_names(1:3, c("a", "b", "c"))
-
-  expect_equal(
-    vec_chop(x, list(1, 2:3)),
-    vec_chop(x, list("a", c("b", "c")))
-  )
-})
-
-test_that("`indices` can use array row names", {
-  x <- array(1:4, c(2, 2), dimnames = list(c("r1", "r2")))
-
-  expect_equal(
-    vec_chop(x, list("r1")),
-    vec_chop(x, list(1))
-  )
-})
-
-test_that("`indices` cannot use data frame row names", {
-  df <- data.frame(x = 1, row.names = "r1")
-  expect_error(vec_chop(df, list("r1")), "Can't use character")
-})
-
-test_that("fallback method with `indices` works", {
-  fctr <- factor(c("a", "b"))
-  indices <- list(1, c(1, 2))
-
-  expect_equal(
-    vec_chop(fctr, indices),
-    map(indices, vec_slice, x = fctr)
-  )
-})
-
-test_that("vec_chop() falls back to `[` for shaped objects with no proxy when indices are provided", {
-  x <- foobar(1)
-  dim(x) <- c(1, 1)
-  result <- vec_chop(x, list(1))[[1]]
-  expect_equal(result, x)
-})
-
 # vec_slice + compact_rep -------------------------------------------------
 
 # `i` is 1-based
@@ -639,6 +498,10 @@ test_that("names are repaired correctly with compact reps and `NA_integer_`", {
   expect_equal(vec_slice_rep(x, NA_integer_, 2L), expect)
 })
 
+test_that("names are recycled correctly with compact reps", {
+  expect_named(vec_slice_rep(c(x = 1L), 1L, 3L), c("x", "x", "x"))
+})
+
 test_that("vec_slice() with compact_reps work with Altrep classes", {
   skip_if(getRversion() < "3.5")
 
@@ -776,71 +639,19 @@ test_that("vec_slice() with compact_seqs work with Altrep classes", {
   expect_equal(vec_slice_seq(x, 1L, 3L), c("foo", "bar", "bar"))
 })
 
-# vec_chop + compact_seq --------------------------------------------------
-
-# `start` is 0-based
-
-test_that("can chop base vectors with compact seqs", {
-  start <- 1L
-  size <- 2L
-  expect_identical(vec_chop_seq(lgl(1, 0, 1), start, size), list(lgl(0, 1)))
-  expect_identical(vec_chop_seq(int(1, 2, 3), start, size), list(int(2, 3)))
-  expect_identical(vec_chop_seq(dbl(1, 2, 3), start, size), list(dbl(2, 3)))
-  expect_identical(vec_chop_seq(cpl(1, 2, 3), start, size), list(cpl(2, 3)))
-  expect_identical(vec_chop_seq(chr("1", "2", "3"), start, size), list(chr("2", "3")))
-  expect_identical(vec_chop_seq(bytes(1, 2, 3), start, size), list(bytes(2, 3)))
-  expect_identical(vec_chop_seq(list(1, 2, 3), start, size), list(list(2, 3)))
+test_that("vec_slice() handles symbols and OO objects", {
+  expect_identical(vec_slice(c(a = 1, b = 2), quote(b)), c(b = 2))
+  expect_identical(vec_slice(c(a = 1, b = 2), factor("b")), c(b = 2))
+  expect_error(vec_slice(c(a = 1, b = 2), foobar("b")), class = "vctrs_error_subscript_type")
 })
 
-test_that("can chop with a decreasing compact seq", {
-  expect_equal(vec_chop_seq(int(1, 2, 3), 1L, 2L, FALSE), list(int(2, 1)))
-})
-
-test_that("can chop with multiple compact seqs", {
-  start <- c(1L, 0L)
-  size <- c(1L, 3L)
-
-  expect_equal(
-    vec_chop_seq(int(1, 2, 3), start, size),
-    list(int(2), int(1, 2, 3))
+test_that("vec_init() handles names in columns", {
+  expect_identical(
+    vec_init(data_frame(x = c(a = 1, b = 2)))$x,
+    named(na_dbl)
+  )
+  expect_identical(
+    vec_init(data_frame(x = c(1, 2)))$x,
+    na_dbl
   )
 })
-
-test_that("can chop S3 objects using the fallback method with compact seqs", {
-  x <- factor(c("a", "b", "c", "d"))
-  expect_equal(vec_chop_seq(x, 0L, 0L), list(vec_slice(x, integer())))
-  expect_equal(vec_chop_seq(x, 0L, 1L), list(vec_slice(x, 1L)))
-  expect_equal(vec_chop_seq(x, 2L, 2L), list(vec_slice(x, 3:4)))
-})
-
-test_that("vec_slice() works with Altrep classes with custom extract methods", {
-  skip_if(getRversion() < "3.5")
-
-  x <- .Call(vctrs_rle, c(foo = 10L, bar = 5L))
-
-  idx <- c(9, 10, 11)
-  expect_equal(vec_slice(x, idx), c("foo", "foo", "bar"))
-})
-
-test_that("slice has informative error messages", {
-  verify_output(test_path("error", "test-slice.txt"), {
-    "# Unnamed vector with character subscript"
-    vec_slice(1:3, letters[1])
-
-    "# Negative subscripts are checked"
-    vec_slice(1:3, -c(1L, NA))
-    vec_slice(1:3, c(-1L, 1L))
-
-    "# oob error messages are properly constructed"
-    vec_slice(c(bar = 1), "foo")
-    "Multiple OOB indices"
-    vec_slice(letters, c(100, 1000))
-    vec_slice(letters, c(1, 100:103, 2, 104:110))
-    vec_slice(set_names(letters), c("foo", "bar"))
-    vec_slice(set_names(letters), toupper(letters))
-
-    "# Can't index beyond the end of a vector"
-    vec_slice(1:2, 3L)
-    vec_slice(1:2, -3L)
-  })
-})
